@@ -88,6 +88,7 @@ class Config(piupartslib.conf.Config):
                 "tmpdir": None,
                 "distro": None,
                 "area": None,
+                "components": None,
                 "chroot-tgz": None,
                 "upgrade-test-distros": None,
                 "basetgz-directory": ".",
@@ -123,37 +124,37 @@ def sighup_handler(signum, frame):
 class MasterIsBusy(Exception):
 
     def __init__(self):
-        self.args = "Master is busy, retry later"
+        self.args = "Master is busy, retry later",
 
 
 class MasterNotOK(Exception):
 
     def __init__(self):
-        self.args = "Master did not respond with 'ok'"
+        self.args = "Master did not respond with 'ok'",
 
 
 class MasterDidNotGreet(Exception):
 
     def __init__(self):
-        self.args = "Master did not start with 'hello'"
+        self.args = "Master did not start with 'hello'",
 
 
 class MasterCommunicationFailed(Exception):
 
     def __init__(self):
-        self.args = "Communication with master failed"
+        self.args = "Communication with master failed",
 
 
 class MasterIsCrazy(Exception):
 
     def __init__(self):
-        self.args = "Master said something unexpected"
+        self.args = "Master said something unexpected",
 
 
 class MasterCantRecycle(Exception):
 
     def __init__(self):
-        self.args = "Master has nothing to recycle"
+        self.args = "Master has nothing to recycle",
 
 
 class Slave:
@@ -412,8 +413,24 @@ class Section:
         oldcwd = os.getcwd()
         os.chdir(self._slave_directory)
 
-        create_or_replace_chroot_tgz(self._config, self._get_tarball(),
-                                     self._config.get_start_distro())
+        tgz = self._get_tarball()
+        max_tgz_age = int(self._config["max-tgz-age"])
+        min_tgz_retry_delay = int(self._config["min-tgz-retry-delay"])
+        needs_update = not os.path.exists(tgz)
+        if not needs_update and max_tgz_age > 0:
+            # tgz exists and age is limited, so check age
+            now = time.time()
+            age = now - os.path.getmtime(tgz)
+            logging.info("Check-replace %s: age=%d vs. max=%d" % (tgz, age, max_tgz_age))
+            if age > max_tgz_age:
+                if os.path.exists(tgz + ".log"):
+                    age = now - os.path.getmtime(tgz + ".log")
+                logging.info("Limit-replace %s: last-retry=%d vs. min=%d" % (tgz, age, min_tgz_retry_delay))
+                if age > min_tgz_retry_delay:
+                    needs_update = True
+                    logging.info("%s too old.  Forcing re-creation" % tgz)
+        if needs_update:
+            create_chroot(self._config, tgz, self._config.get_start_distro())
 
         os.chdir(oldcwd)
 
@@ -576,6 +593,7 @@ class Section:
     def _process(self):
         self._slave.close()
 
+        packagenames = set([x[0] for x in self._slave.get_reserved()])
         packages_files = {}
         for distro in [self._config.get_distro()] + self._config.get_distros():
             if distro not in packages_files:
@@ -585,19 +603,26 @@ class Section:
                             self._distro_config.get_packages_urls(
                                 distro,
                                 self._config.get_area(),
-                                self._config.get_arch()))
+                                self._config.get_arch()),
+                            packagenames)
                     packages_files[distro] = pf
                 except IOError:
                     logging.error("failed to fetch packages file for %s" % distro)
                     self._error_wait_until = time.time() + 900
                     return 0
+        del packagenames
 
         test_count = 0
         self._check_tarball()
+        if not os.path.exists(self._get_tarball()):
+            self._error_wait_until = time.time() + 300
         for package_name, version in self._slave.get_reserved():
             self._throttle_if_overloaded()
             if interrupted or got_sighup:
                 break
+            if not os.path.exists(self._get_tarball()):
+                logging.error("Missing chroot-tgz %s" % self._get_tarball())
+                break
             test_count += 1
             self._test_package(package_name, version, packages_files)
             self._slave.forget_reserved(package_name, version)
@@ -626,7 +651,10 @@ class Section:
         if "http_proxy" in os.environ:
             command.extend(["--proxy", os.environ["http_proxy"]])
         if self._config["mirror"]:
-            command.extend(["--mirror", self._config["mirror"]])
+            mirror = self._config["mirror"]
+            if self._config["components"]:
+                mirror += " " + self._config["components"]
+            command.extend(["--mirror", mirror])
         if self._config["tmpdir"]:
             command.extend(["--tmpdir", self._config["tmpdir"]])
         command.extend(["--arch", self._config.get_arch()])
@@ -704,7 +732,7 @@ class Section:
             subdir = "untestable"
 
         if ret == 0:
-            output.write("Executing: %s\n" % " ".join(command))
+            output.write("Executing: %s\n" % " ".join(quote_spaces(command)))
             ret, f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN)
             if not f or f[-1] != '\n':
                 f += '\n'
@@ -732,6 +760,10 @@ def log_name(package, version):
     return "%s_%s.log" % (package, version)
 
 
+def quote_spaces(vlist):
+    return ["'%s'" % x if ' ' in x else x for x in vlist]
+
+
 def run_test_with_timeout(cmd, maxwait, kill_all=True):
 
     def terminate_subprocess(p, kill_all):
@@ -771,7 +803,7 @@ def run_test_with_timeout(cmd, maxwait, kill_all=True):
                 except OSError:
                     pass
 
-    logging.debug("Executing: %s" % " ".join(cmd))
+    logging.debug("Executing: %s" % " ".join(quote_spaces(cmd)))
 
     stdout = ""
     p = subprocess.Popen(cmd, preexec_fn=os.setpgrp,
@@ -802,63 +834,46 @@ def run_test_with_timeout(cmd, maxwait, kill_all=True):
 
 
 def create_chroot(config, tarball, distro):
-    output_name = tarball + ".log"
-    logging.debug("Opening log file %s" % output_name)
-    logging.info("Creating new tarball %s" % tarball)
     command = config["piuparts-command"].split()
     if config["piuparts-flags"]:
         command.extend(config["piuparts-flags"].split())
     if "http_proxy" in os.environ:
         command.extend(["--proxy", os.environ["http_proxy"]])
     if config["mirror"]:
-        command.extend(["--mirror", config["mirror"]])
+        mirror = config["mirror"]
+        if config["components"]:
+            mirror += " " + config["components"]
+        command.extend(["--mirror", mirror])
     if config["tmpdir"]:
         command.extend(["--tmpdir", config["tmpdir"]])
     command.extend(["--arch", config.get_arch()])
     command.extend(["-d", distro])
     command.extend(["-s", tarball + ".new"])
     command.extend(["--apt", "dpkg"])
-    output = file(output_name, "w")
-    output.write(time.strftime("Start: %Y-%m-%d %H:%M:%S %Z\n\n",
-                               time.gmtime()))
-    output.write("Executing: " + " ".join(command) + "\n\n")
-    logging.debug("Executing: " + " ".join(command))
-    p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
-    for line in p.stdout:
-        output.write(line)
-        logging.debug(">> " + line.rstrip())
-    p.wait()
-    output.write(time.strftime("\nEnd: %Y-%m-%d %H:%M:%S %Z\n",
-                               time.gmtime()))
-    output.close()
-    if os.path.exists(tarball + ".new"):
-        os.rename(tarball + ".new", tarball)
 
-def create_or_replace_chroot_tgz(config, tgz, distro):
-    max_tgz_age = int(config["max-tgz-age"])
-    min_tgz_retry_delay = int(config["min-tgz-retry-delay"])
-    forced = 0
-    if os.path.exists(tgz) and max_tgz_age > 0:
-        now = time.time()
-        statobj = os.stat(tgz)
-        # stat.ST_MTIME is actually time file was initially created
-        age = now - statobj[stat.ST_MTIME]
-        logging.info("Check-replace %s: age=%d vs. max=%d" % (tgz, age, max_tgz_age))
-        if age > max_tgz_age:
-            logging.info("Limit-replace %s: last-retry=%d vs. min=%d" % (tgz, now - statobj[stat.ST_CTIME], min_tgz_retry_delay))
-            # stat.ST_CTIME is time created OR last renamed
-            if min_tgz_retry_delay is None or now - statobj[stat.ST_CTIME] > min_tgz_retry_delay:
-                os.rename(tgz, tgz + ".old")
-                forced = 1
-                logging.info("%s too old.  Renaming to force re-creation" % tgz)
-    if not os.path.exists(tgz):
-        create_chroot(config, tgz, distro)
-        if forced:
-            if not os.path.exists(tgz):
-                os.rename(tgz + ".old", tgz)
-                logging.info("Failed to create ... reverting to old %s" % tgz)
+    output_name = tarball + ".log"
+    with open(output_name, "we") as output:
+        try:
+            fcntl.flock(output, fcntl.LOCK_EX | fcntl.LOCK_NB)
+        except IOError:
+            logging.info("Creation of tarball %s already in progress." % tarball)
+        else:
+            logging.info("Creating new tarball %s" % tarball)
+            output.write(time.strftime("Start: %Y-%m-%d %H:%M:%S %Z\n\n",
+                                       time.gmtime()))
+            output.write("Executing: " + " ".join(quote_spaces(command)) + "\n\n")
+            logging.debug("Executing: " + " ".join(quote_spaces(command)))
+            p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+            for line in p.stdout:
+                output.write(line)
+                logging.debug(">> " + line.rstrip())
+            p.wait()
+            output.write(time.strftime("\nEnd: %Y-%m-%d %H:%M:%S %Z\n",
+                                       time.gmtime()))
+            if os.path.exists(tarball + ".new"):
+                os.rename(tarball + ".new", tarball)
             else:
-                os.unlink(tgz + ".old")
+                logging.error("Tarball creation failed, see %s" % output_name)
 
 
 def create_file(filename, contents):
